
(* forf frac -G; prints the abstract tts by Prn *)

functor EmitAbstract (Prn : PRN) = struct

open Prn;

fun newline n = if n<0 then () else (print "\n"; repeat n print " ");

fun print_Fiacre (Program sl) =
	List.app (fn (_,dl,m) => (List.app print_Decl dl; print_Decl m)) sl
and print_Decl d =
        case d of
	    TypeDecl d => print_TypeDecl 0 d
       (* | ChannelDecl d => print_ChannelDecl 0 d *)
	  | FunctionDecl d => print_Function 0 d
	  | ExternDecl d => print_Extern 0 d
	  | TaskDecl d => print_Task 0 d
	  | EventDecl d => print_Event 0 d			    
	  | ProcessDecl d => print_Process 0 d
	  | ComponentDecl d => print_Component 0 d
	  | Main (m,low) => (newline 0; print m;
			     case low of [] => () | _ => (print " > "; print_list (""," | ",")") print low))
	  | PropertyDecl d => print_Property 0 d
	  | Assert ("",_,_) => ()
	  | Assert (p,tt,ff) => (newline 0; print "assert "; print p;
				 case tt of SOME s => (print " -true "; print s) | _ => ();
		                 case ff of SOME s => (print " -false "; print s) | _ => ();
				 print "\n")
and print_TypeDecl c (a,b) =
	(newline c; print "type "; print a; print " is "; print_Type b; newline 0)
(*
and print_ChannelDecl c (l,a,b) =
	(newline c; print "channel "; print a; print " is "; print_Channel b; newline 0)
*)
and print_Type ty = (case ty of
	  BoolType => print "bool"
	| NatType => print "nat"
	| IntType => print "int"
	| CharType => print "char"
	| UnsignedCharType => print "unsigned char"
	| ShortType => print "short"
	| UnsignedShortType => print "unsigned short"
	| NamedType a => print a
	| IntervalType(a,b) => (printi a; print ".."; printi b)
	| EnumType a => print_list ("enum "," | "," end") print a  (* no enums shown *)
	| UnionType a => print_list ("union "," | "," end") (fn (xl,NONE) => print xl | (xl,SOME y) => (print xl; print " of "; print_Type y)) a
	| RecordType a 	=> print_list ("record ",", "," end") (fn (xl,y) => (print xl; print ":"; print_Type y)) a
	| ArrayType (e,b) => (print "array "; printi e; print " of "; print_Type b)
	| QueueType (e,b) => (print "queue "; printi e; print " of "; print_Type b))
and print_Channel ch =
        print_list (""," # ","") print_Type ch
and print_Profile tyl =
        case tyl of
	    [] => print "sync"
	  | _  => print_Channel tyl
and print_Function i (f,args,ty,vars,body) =
	(newline i; print "function "; print f;
	 print_list (" (",", ",")") (fn (v,ty) => (print v; print ": "; print_Type ty)) args;
	 print " : "; print_Type ty;
	 print " is ";
	 if null vars then () else
	 (newline (i+4); print_list ("var ",", ","") (fn (s,ty,v) => (print s; print ": "; print_Type ty; print ":= "; print_Exp v)) vars);
	 newline (i+4); print "begin";
	 print_Statement (i+6) body;
	 newline (i+4); print "end";
	 newline 0)
and print_Extern i (f,argtypes,ty) =
	(newline 0; print "extern "; print_Type ty; print f;
	 print_list ("(",", ",");") print_Type argtypes;
	 newline 0)
and print_Task i (t,sym,argtypes,ty) =
	(newline 0; print "/* fiacre task "; print t; print ", cname "; print sym; print " */";
	 (* needed ? *)
	 newline 0; print "extern "; print_Type ty; print sym; 
	 print_list ("(",", ",");") print_Type argtypes;
	 (* t_input *)
	 let val next = let val N = ref 0 in fn () => (N := 1 + !N; Int.toString (!N)) end
	 in List.app (fn ty => (newline 0; print "extern "; print_Type ty; print (" "^sym^"_input"^next()); print ";")) argtypes
	 end;
	 (* t_output *)
	 newline 0; print "extern "; print_Type ty; print (" "^sym^"_output"); print ";";
	 (* t_ready, t_terminate *)
	 newline 0; print "extern int "; print (sym^"_ready"); print ";";
	 newline 0; print "extern int "; print (sym^"_terminate"); print ";";
	 newline 0)
and print_Event i (e,sym,types) =
	(newline 0; print "/* fiacre event "; print e; print ", cname "; print sym; print " */";
	 (* e_ready *)
	 newline 0; print "extern int "; print (sym^"_ready"); print ";";
	 (* e_values *)
	 let val next = let val N = ref 0 in fn () => (N := 1 + !N; Int.toString (!N)) end
	 in List.app (fn ty => (newline 0; print "extern "; print_Type ty; print (" "^sym^"_value"^next()); print ";")) types
	 end;
	 newline 0)
and print_Process i (a,b,d,sd,rk,k,f,h,g,init,j) =
	(newline 0; print "process "; print a;
	 case b of [] => () | _ => (newline 0; newline (i+4); print_list ("[",", ","]") print_PortDecl b);
	 case d of [] => () | _ => (newline 0; newline (i+4); print_list ("{",", ","}") print_VarParam d);
	 newline (i+4); print "is";
	 newline 0; newline (i+4); print_StatesDecl sd;
	 case rk of [] => () | _ => (newline 0; newline (i+4); print_list ("const ",", ","") print_VarDecl rk);
	 case k of [] => () | _ => (newline 0; newline (i+4); print_list ("tmp ",", ","") print_VarDecl k);
	 case f of [] => () | _ => (newline 0; newline (i+4); print_list ("var ",", ","") print_VarDecl f);
	 case h of [] => ()
		 | h::t => (newline 0; newline (i+4); print "// port "; print_locPortDecl h;
			    List.app (fn p => (print ","; newline (i+4); print "//      "; print_locPortDecl p)) t);
	 case g of [] => () | _ => (newline 0; newline (i+4); print_priorities g);
	 case init of StNull => () | StTo [] => () | _ => (newline 0; newline (i+4); print "init "; print_Statement (i+4) init);
	 app (print_Transition (i+4)) j;
	 newline 0)
and print_Component i (a,b,d,rk,k,e,f,g,init,h) =
	(newline 0; print "component "; print a;
	 case b of [] => () | _ => (newline (i+4); print_list ("[",", ","]") print_PortDecl b);
	 case d of [] => () | _ => (newline (i+4); print_list ("{",", ","}") print_VarParam d);
	 newline (i+4); print "is";
	 case rk of [] => () | _ => (newline (i+4); print_list ("const ",", ","") print_VarDecl rk);
	 case k of [] => () | _ => (newline (i+4); print_list ("tmp ",", ","") print_VarDecl k);
	 case e of [] => () | _ => (newline (i+4); print_list ("var ",", ","") print_VarDecl e);
	 case f of [] => ()
		 | h::t => (newline (i+4); print "port "; print_locPortDecl h;
			    List.app (fn p => (print ","; newline (i+9); print_locPortDecl p)) t);
	 case g of [] => () | _ => (newline 0; newline (i+4); print_priorities g);
	 case init of StNull => () | StTo [] => () | _ => (newline 0; newline (i+4); print "init "; print_Statement (i+4) init);
	 print_Composition (i+4) h false;
	 newline 0)
and print_priorities [] = ()
  | print_priorities l = print_list ("priority ",", ","")
				    (fn (l1,l2) => (print_list (""," ","") print l1; print " > "; print_list (""," ","") print l2)) 
				    l
and print_locPortDecl (name,ch,pb,(left,right)) =
        (print name; print ": "; print_Profile ch; print_TimeConstraint (left,right);
	 case pb of SOME p => (print " is "; print_Probe p) | _ => ())
and print_Composition i c x = (case c of (* x true if starts at beginning of line *)
	  ParComp cl  => (if x then () else newline i; print "par";
			  let val (pl,h)::t = cl
			      val no = ref false
			  in newline (i+3);
			     case pl of [] => no:=true | _ => (print_list ("",",","") print pl; print " -> ");
			     print_Composition (i+3) h (!no);
			     List.app (fn (pl,c) =>
					  (newline i; print "|| ";
					   case pl of [] => no:=true | _ => (print_list ("",",","") print pl; print " -> ");
					   print_Composition (i+3) c (!no))) t
			  end;
			  newline i; print "end")
	| InstanceComp c => print_InstanceComp i c
	| PrioSpec (prl,c) => (if x then () else newline i;
			       print "prio "; print_list ("",", ","") (fn (x,y) => (print x; print " > "; print y)) prl;
			       print_Composition i c false)
	| RenSpec (rnl,c) => (if x then () else newline i;
			      print "renaming "; print_list ("",", ","") (fn (x,y) => (print x; print " -> "; print y)) rnl;
			      print_Composition i c false))
and print_InstanceComp i (id,pl,al) =
	(print id;
	 if null pl then () else print_list (" [",",","]") print pl;
	 if null al then () else print_list (" (",",",")") print_Exp al)
and print_PortDecl (name,ch) =
	(print name; print ":"; print_Profile ch)
and print_arg (ValArg s) = print s | print_arg (RefArg s) = print s
and print_VarParam (name,ty) =
	(print_arg name; print ":"; print_Type ty)
and print_VarDecl (a,b,c) =
	(print a; print ": "; print_Type b; print " := "; print_Exp c)
and print_StatesDecl a =
	(print_list ("states ",", ","") print a)
and print_TimeConstraint (left,right) =
        if case (left,right) of (Closed d,Infinite) => d <= 0.0 | _ => false then () else
	(print " in "; print_Interval (left,right))
and print_Interval (left,right) =
    let fun printb b = print (Int.toString (Real.floor (b * (pow (10.0,!DIGITS)))))
    in case left of Open b => (print "]"; printb b)
		  | Closed b => (print "["; printb b);
       print ",";
       case right of Open b => (printb b; print "[")
		   | Closed b => (printb b; print "]")
		   | Infinite => print "...["
    end
and print_Transition i ((tags,lab,atags),b,c,bnd,prl) =
	(newline 0; newline i; print "Trans::";
	 print (String.concatWith "__" tags);
	 case lab of "" => () | _ => print (": " ^ lab);
	 case atags of [] => () | _ => (print " & "; print (String.concatWith " " atags));
	 newline (i+2); print "from "; print_list ("",", ","") print b; print " "; print_Statement (i+3) c;
	 newline (i+2); print "in "; print_Interval bnd;
	 case prl of [] => () | _ => (newline (i+2); print "> "; print_list (""," ","") print prl))
and print_Statement e st = case st of
	  StNull => (newline e; print "null")
	| StGuard a => (newline e; print "on "; print_Exp a)
	| StCond(a,b,c) =>
		(newline e; print "if "; print_Exp a; print " then "; print_Statement (e+3) b;
		 newline e; print " else ";  print_Statement (e+3) c;
		 newline e; print " end /* if */")
	| StSelect (a::t) =>
	        let fun print_alternant e (a::t) =
		  (print_Statement (e+3) a;
		   List.app (fn y => (newline e; print "[]"; print_Statement (e+3) y)) t);
	        in newline e; print "select";
		   print_alternant e a;
		   List.app (fn y => (newline e; print "unless"; print_alternant e y)) t;
		   newline e; print "end /* select */"
                 end
	| StSequence(a,b) => (print_Statement e a; print "; "; print_Statement e b)
	| StSignal a => (newline e; print a)
	| StInput (a,r,b) =>
		(newline e; print a; print_pdetails r; print "? "; print_list ("",",","") print_Patt b)
	| StInputList (a,b) =>
		(print a; print_list ("[","-","]") (fn y => print_list ("",",","") print_Exp y) b)
  	| StOutput (a,r,b) =>
		(newline e; print a; print_pdetails r; print "! "; print_list ("",", ","") print_Exp b)
	| StAssign (a,b) =>
		(newline e; print_Patt a; print " := "; print_Exp b)
	| StWhile(a,b) => (newline e; print "while "; print_Exp a; print " do "; 
			     print_Statement (e+3) b; 
			     newline e; print " end /* while */")
	| StForeach(a,b) => (newline e; print "foreach "; print_Exp a; print " do "; 
			     print_Statement (e+3) b; 
			     newline e; print " end /* foreach */")
	| StTo a => (newline e; print "to "; print_list ("",", ","") print a)  (* ### loop deduced from src list ### *)
	| StBlock (dl,s) => (newline e; print "block {"; 
			     map (fn (v,ty) => (newline (e+2); print v; print " : "; print_Type ty; print "; ")) dl;
			     print_Statement (e+2) s;
			     newline e; print "} /* block */")
	| StReturn exp => (newline e; print "return "; print_Exp exp)
	| StFailwith (MatchError f) => (newline e; print "failwith (MatchError \""; printi f; print "\")")
	| _ => failwith "unexpected statement in print_Statement"
and print_Patt e =
        let val d = !DETAILS in DETAILS := false; print_Exp e; DETAILS := d end
and print_Exp e = case e of
	  VarExp (s,r)          => (print s; print_details r)
	| CVarExp (s,r)         => (print s; print_details r)
	| EnumExp (s,r)         => (print s; print_details r)
	| ConExp (s,r)          => (print s; print_details r)
	| ArrayAccessExp(a,b,_) => (print_Exp a; print "["; print_Exp b; print "]")
 	| RecordAccessExp(a,b,_) => (print_Exp a; print "."; print b)
	| IntExp (n,r)          => (printi n; print_details r)
	| BoolExp b             => print (Bool.toString b)
	| InfixExp(x,r,y,z)     => (print "("; print_Exp y; print " "; print_Infix x; print_details r; print " "; print_Exp z; print ")")
	| BinopExp(x,r,y,z)	=> (print_Binop x; print_details r; print "("; print_Exp y; print ","; print_Exp z; print ")")
	| UnopExp(a,r,b)        => (print_Unop a; print_details r; print "("; print_Exp b; print ")")
	| ConstrExp(c,e2,r)     => (print c; print "("; print_Exp e2; print_details r; print ")")
	| AppExp(f,el,_,r)      => (print f; print_list (" (",",",")") print_Exp el; print_details r)
	| CallExp(f,el,_,r)     => (print ("\""^f^"\""); print_list (" (",",",")") print_Exp el; print_details r)
	| RecordExp (fl,r)      => let val b = !DETAILS
				   in DETAILS := false; print_list ("{",",","}") (fn (f,l) => (print f; print "="; print_Exp l)) fl;
				      DETAILS := b; print_details r
				   end
	| ArrayExp (el,r)       => let val b = !DETAILS
				   in DETAILS := false; print_list ("[",",","]") print_Exp el; DETAILS := b; print_details r
			           end
	| QueueExp (el,r)   	=> (print_list ("{|",",","|}") print_Exp el; print_details r)
	| CondExp (c,e1,e2)     => (print "("; print_Exp c; print ") ? ("; print_Exp e1; print ") : ("; print_Exp e2; print ")")
	| RefExp (v,r)          => (print "&"; print v; print_details r)
        | LetExp(s,e)           => (print "({"; print_Statement (~1000) s; print "; "; print_Exp e; print "})")
	| DestrExp (c,e)        => (print "Destr ("; print_Exp e; print ")")
	| MatchesExp (c,e)      => (print "Matches ("; print c; print ","; print_Exp e; print ")")
	| AnyExp []             => print "any"
	| AnyExp _              => failwith "unexpected 'any' pattern in send"
(* 	| _                     => failwith "unexpected exp in print_Exp" *)
and print_details ty =
	if !DETAILS then (print "::"; print_Type ty) else ()
and print_rdetails tyl =
	if !DETAILS then (print "::"; print_list (""," * ","") print_Type tyl) else ()
and print_pdetails ty =
	if !DETAILS then (print "::"; print_Channel ty) else ()
and print_Infix ope = print (case ope of
	  AND => "and"	| OR  => "or"	| LE  => "<="	| LT  => "<"	| GT  => ">"	| GE  => ">="
	| EQ  => "="	| NE  => "<>"	| ADD => "+"	| SUB => "-"	| MOD => "%"	| MUL => "*"	| DIV => "/")
and print_Binop ENQUEUE = print "enqueue" | print_Binop APPEND = print "append"
and print_Unop ope = print (case ope of
	  COERCE => "$" | NOT => "not" | ASSERT _ => "assert" | FULL => "full" | EMPTY => "empty" | DEQUEUE => "dequeue"
	| PLUS => "+" | MINUS => "-" | FIRST => "first" | LENGTH => "length")
and print_Probe pb = (case pb of
          StateProbe (_,s) => (print "state "; print s)
        | ValueProbe (_,e) => (print "value "; print_Exp e)
        | EventProbe (_,r) => (print "event "; print r)
        | TagProbe (_,r)   => (print "tag "; print r)
        | EnterProbe r => (print "enter ("; print_Probe r; print ")")
        | LeaveProbe r => (print "leave ("; print_Probe r; print ")")
	| NotProbe r => (print "not ("; print_Probe r; print ")")
	| AndProbe (r1,r2) => (print_Probe r1; print " and "; print_Probe r2)
	| OrProbe (r1,r2) => (print_Probe r1; print " or "; print_Probe r2)
	| ImpProbe (r1,r2) => (print_Probe r1; print " => "; print_Probe r2))
and print_Bound bnd =
    case bnd of
	Closed b => print (Int.toString (Real.floor (b * (pow (10.0,!DIGITS)))))
      | Open b => print (Int.toString (Real.floor (b * (pow (10.0,!DIGITS)))))
      | _ => print "..."
and print_lasting (b as Closed _) = (print " lasting "; print_Bound b)
  | print_lasting _ = print "?"
and print_duration (SOME (b as Closed _)) = (print " for "; print_Bound b)
  | print_duration (SOME _) = (print " for "; print "?")
  | print_duration _ = ()
and print_within (SOME i) = (print " within "; print_Interval i)
  | print_within _ = ()
and print_Prop p = (case p of
         DeadlockFreeProp => print "deadlockfree"
       | InfinitelyOftenProp p => (print "infinitelyoften "; print_Probe p)
       | MortalProp p => (print "mortal "; print_Probe p)
       | LiveProp p => (print "live "; print_Probe p)
       | PresentProp (p,i) => (print "present "; print_Probe p; print_within i)
       | PresentLastingProp (p,d) => (print "present "; print_Probe p; print_lasting d)
       | PresentAfterProp (p1,p2,i) => (print "present "; print_Probe p1; print " after "; print_Probe p2; print_within i)
       | PresentBeforeProp (p1,p2,i) => (print "present "; print_Probe p1; print " before "; print_Probe p2; print_within i)
       | PresentBetweenProp (p1,p2,p3,i) => (print "present "; print_Probe p1; print " between "; print_Probe p2; print " and "; print_Probe p3; print_within i)
       | PresentUntilProp (p1,p2,p3,i) => (print "present "; print_Probe p1; print " after "; print_Probe p2; print " until "; print_Probe p3; print_within i)
       | AbsentProp (p,i) => (print "absent "; print_Probe p; print_within i)
       | AbsentLastingProp (p,d) => (print "absent "; print_Probe p; print_lasting d)
       | AbsentAfterProp (p1,p2,i) => (print "absent "; print_Probe p1; print " after "; print_Probe p2; print_within i)
       | AbsentBeforeProp (p1,p2,d) => (print "absent "; print_Probe p1; print " before "; print_Probe p2; print_duration d)
       | AbsentBetweenProp (p1,p2,p3,i) => (print "absent "; print_Probe p1; print " between "; print_Probe p2; print " and "; print_Probe p3; print_within i)
       | AbsentUntilProp (p1,p2,p3,i) => (print "absent "; print_Probe p1; print " after "; print_Probe p2; print " until "; print_Probe p3; print_within i)
       | LeadstoProp (p1,p2,i) => (print_Probe p1; print " leadsto "; print_Probe p2; print_within i)
       | LeadstoBeforeProp (p1,p2,p3,i) => (print_Probe p1; print " leadsto "; print_Probe p2; print " before "; print_Probe p3; print_within i)
       | LeadstoAfterProp (p1,p2,p3,i) =>  (print_Probe p1; print " leadsto "; print_Probe p2; print " after "; print_Probe p3; print_within i)
       | LeadstoBetweenProp (p1,p2,p3,p4,i) => (print_Probe p1; print " leadsto "; print_Probe p2; print " between "; print_Probe p3; print " and "; print_Probe p4; print_within i)
       | LeadstoUntilProp (p1,p2,p3,p4,i) => (print_Probe p1; print " leadsto "; print_Probe p2; print " after "; print_Probe p3; print " until "; print_Probe p4; print_within i)
       | AlwaysProp (p,i) => (print "always "; print_Probe p; print_within i)
       | AlwaysLastingProp (p,b) => (print "always "; print_Probe p; print_lasting b)
       | AlwaysAfterProp (p1,p2,b) => (print "always "; print_Probe p1; print " after "; print_Probe p2; print_duration b)
       | AlwaysBeforeProp (p1,p2,b) => (print "always "; print_Probe p1; print " before "; print_Probe p2; print_duration b)
       | AlwaysBetweenProp (p1,p2,p3,i) => (print "always "; print_Probe p1; print " between "; print_Probe p2; print " and "; print_Probe p3; print_within i)
       | AlwaysUntilProp (p1,p2,p3,i) => (print "always "; print_Probe p1; print " after "; print_Probe p2; print " until "; print_Probe p3; print_within i)
       | PrecedesProp (p1,p2,b) => (print_Probe p1; print " precedes "; print_Probe p2; print_duration b)
       | PrecedesBeforeProp (p1,p2,p3,b) => (print_Probe p1; print " precedes "; print_Probe p2; print " before "; print_Probe p3; print_duration b)
       | PrecedesAfterProp (p1,p2,p3,b) => (print_Probe p1; print " precedes "; print_Probe p2; print " after "; print_Probe p3; print_duration b)
       | PrecedesBetweenProp (p1,p2,p3,p4,b) => (print_Probe p1; print " precedes "; print_Probe p2; print " between "; print_Probe p3; print " and "; print_Probe p4; print_duration b)
       | PrecedesUntilProp (p1,p2,p3,p4,b) => (print_Probe p1; print " precedes "; print_Probe p2; print " after "; print_Probe p3; print " until "; print_Probe p4; print_duration b)
       | AndProp (p1,p2) => (print_Prop p1; print " and "; print_Prop p2)
       | OrProp (p1,p2) => (print_Prop p1; print " or "; print_Prop p2)
       | ImpProp (p1,p2) => (print_Prop p1; print " => "; print_Prop p2)
       | NotProp p => (print "not "; print_Prop p)
       | BoxProp p => (print "[] "; print_Prop p)
       | DiaProp p => (print "<> "; print_Prop p)
       | NextProp p => (print "() "; print_Prop p)
       | UntilProp (p1,p2) => (print_Prop p1; print " until "; print_Prop p2)
       | ReleaseProp (p1,p2) => (print_Prop p1; print " release "; print_Prop p2)
       | ProbeProp p => (print "["; print_Probe p; print "]")
       | VarProp n => print n)
and print_Property _ (s,p) =
    (newline 0; print "property "; print s; print " is "; print_Prop p; newline 0);

fun emit s0 = (print_Fiacre (Name.nameTypes s0); newline 0; flush ()) handle BYE => raise BYE | _ => failwith "emit tts";

end; (* functor EmitAbstract *)

	
structure OutEmitAbstract = EmitAbstract(OutPrint);
structure ErrEmitAbstract = EmitAbstract(ErrPrint);


