(* prints the Hippo C code by Prn *)
structure HippoCEmit = struct

(* SharedCEmit brings
references;
    xtr  (for hippo only)
    TYPES,TEMP,IMPURE,PURE
functions:
    newline
    getBoundType
    new
    print_Type
    print_ExternDecl
    print_compareFunc
    print_sprintFunc
    print_VarDecl
    assigns
    and assignse
    and print_Exp
    print_Statement
    print_FunctionDecl
    print_Initializer
    print_preFunc
 *)
open SharedCEmit;

(* now in SharedCEmit:
val xtr = ref NONE;  (* temporary for recording extern *)
*)
val XTR = ref(Array.array(0,NONE));  (* extern array for computing externsets *)
val PRE = ref(Array.array(0,"0"));
val ACT = ref(Array.array(0,"0"));     

(* ### SEE: what type should be printed in code generated if profile has several types ? ### *)
(* ### currently prints void if profile is sync, and raises error if several types ### *)
fun print_Profile tyl sym a =
    let fun printn "" = () | printn a = (print " "; print a)
    in case tyl of
	    [] => (print "void"; printn a)
	  | [ty] => print_Type ty a
	  | _  => comperror ("events with general profiles currently not supported (" ^ sym ^ ")")
    end;

(* printing header file *)
fun print_Headers (Program ((_,dl,_)::_)) =
    let fun print_dec d =
	    case d of
	 	TypeDecl (a,ty) => (newline 0; newline 0; print_Type ty a; print ";")
	      |	TaskDecl (t,sym,argtypes,ty) => (newline 0; newline 0; print_Type ty ""; print " "; print sym; 
						 print_list ("(",", ",")") (fn ty => print_Type ty "") argtypes; print ";")
	      | EventDecl (e,sym,ty) => (newline 0; newline 0; print_Profile ty sym ""; print " "; print sym; print "();")
	      | _ => ()
    in newline 0; print "#ifndef _APP_FUNCTIONS_\n#define _APP_FUNCTIONS_\n/* fiacre headers */";
       newline 0;(* print "#include \"fiacre.h\"";*)
       print "void initial();";
       List.app print_dec dl;
       newline 0;
       newline 0;
       print "// fiacre type bindings\n";
       (* note: no "product" channels supported by hippo yet, silently skipped here, if any *)
       List.app (fn TypeDecl (n,ty)      => (print "typedef "; print_Type ty (n ^"_c;\n"))
		  | ChannelDecl (n,[ty]) => (print "typedef "; print_Type ty (n ^"_c;\n"))
		  |  _ => ()) (!TYCH_DECLS);
       newline 0;
       print "#endif"
    end;
				
fun print_TaskDecl (TaskDecl (t,sym,argtypes,ty)) =
    ((* also adds extern variables to TEMP *)
     newline 0; print "/* fiacre task "; print t; print ", cname "; print sym; print " */";
     (* needed ? *)
     (*newline 0; print "extern "; print_Type ty ""; print " "; print sym; 
     print_list ("(",", ",")") (fn ty => print_Type ty "") argtypes; print ";";*)
     (* t_input *)
     let val next = let val N = ref 0 in fn () => (N := 1 + !N; Int.toString (!N)) end
     in List.app (fn ty => (newline 0; print_Type ty ""; print (" "^sym^"_input_"^next()); print ";"))
		 argtypes
     end;
     (* t_output *)
     newline 0; print_Type ty ""; print (" "^sym^"_output"); print ";";
     (* t_start, t_terminate *)
     newline 0; print "extern int "; print (sym^"_start"); print ";";
     newline 0; print "extern int "; print (sym^"_terminate"); print ";";
    (*wrapper bad code*)
     newline 0; print"void wrapper_"; print sym; print "() {";
     newline 4; print sym; print "_output = "; print sym; print "(";
     let val next = let val N = ref 0 in fn () => (N := 1 + !N; Int.toString (!N)) end
     in List.app (fn (n,ty) => (print (" "^sym^"_input_"^next()); if n<> ((List.length argtypes) - 1) then (print ",") else ()))
		(ListPair.zip(List.tabulate(List.length argtypes,fn z => z), argtypes))
     end;
     print ");"; newline 0; print "}");
    
fun print_EventDecl (EventDecl (e,sym,types)) =
    (newline 0; print "/* fiacre event "; print e; print ", cname "; print sym; print " */";
     (* e declaration*)
     (* newline 0; print "extern "; print_Profile types sym ""; print " "; print sym; print "();";*)
     (* e_ready *)
     newline 0; print "extern int "; print (sym^"_ready"); print ";";
     (* e_values *)
     let val next = let val N = ref 0 in fn () => (N := 1 + !N; Int.toString (!N)) end
     in List.app (fn ty => (newline 0; (fn ty => print_Type ty "") ty; print (" "^sym^"_value_"^next()); print ";")) types
     end;
     (* wrapper*)
     newline 0; print"void wrapper_"; print sym; print "() {";
     let val next = let val N = ref 0 in fn () => (N := 1 + !N; Int.toString (!N)) end
     in List.app (fn ty => (newline 4;  print (" "^sym^"_value_"^next()); print " = "; print sym;print"();")) types
     end;
     newline 0; print "}");

fun sprint_value_hippo spr ftyl =
    (newline 0; print "\nstatic int sprint_value (int sz, char *buff, const struct pmap *pmap) {";
     newline 2; print "struct value *v = &state;";
     newline 2; print "avail = sz - 1;";
     newline 2; print "/* prints state into buff: */";
     newline 2; print "PUTS(pmap->r.l);";
     spr ftyl;
     newline 2; print "PUTS(pmap->r.r);";
     newline 2; print "return (avail<0) ? -1 : (sz - avail - 1);";
     newline 0; print "}";
     newline 0; print "\nint sprint_state (int sz, char *buff) {";
     newline 2; print "return sprint_value (sz, buff, &fcrPmap);";
     newline 0; print "}";
     newline 0; print "\nint sprint_json_state (int sz, char *buff) {";
     newline 2; print "return sprint_value (sz, buff, &jsonPmap);";
     newline 0; print "}");

(* emit type declarations for named types *)
fun print_TypeDecl (TypeDecl (a,ty)) =
    ((* print "\n\n"; print_Type ty a; print ";" *)  (* now printed in header file *)
     print "\n\n/* utilities for structure "; print a; print " */";
     print_compareFunc (a,ty);
     print_sprintFunc sprint_value_hippo (a,ty));

fun print_initialFunc a b = (* a = state variables, b = rtmp variables *)
    (newline 0; newline 0; print "/* initializes temporaries and creates initial state */";
     newline 0; print "void initial() {"; (* //&*)
     (*newline 2; print "struct value start;"; //&*)
     (* initialize temporaries: *)
     if null b then () else (newline 2; print "/* read-only temporaries: */");
     List.app (fn x as (st,_,v) => print_Initializer st v "") b;
     (* initialize state variables: *)
     newline 2; print "/* state variables: */";
     List.app (fn x as (st,_,v) => print_Initializer st v "state.") a; (*//&*)
     (* establishes value probes for initial state *)
     newline 0; print "}";
     newline 0);

fun print_pre_hippo print_guard e tag index dl k =
    (Array.update(!PRE,index,"pre_"^Int.toString(index));  
     newline 0; newline 0; print ("bool pre_"^ Int.toString(index) ^" () {"); (*//&*)
     newline 2; print "/* "; print tag; print " */";
     (* if null k then () else (newline 2; print "/* temporaries */"); *)
     List.app (fn d => (newline 2; print_VarDecl d)) k;
     xtr := NONE;  (* reset hippo task or event *)
     if assignse dl e
     then (* #### improve: make local copies of state variables updated in the block, rather than all #### *)
	 (newline 2;
	  print "struct value new_state = state;";
	  print_guard dl e "new_state.")
     else (newline 2;
	   print "struct value *v = &state;";
	   print_guard dl e "v->"); 
     Array.update(!XTR,index,!xtr);  (* records local hippo task or event *)
     newline 2; print "return 1;"; 
     newline 1; print ("error: return 0;");
     newline 0; print "}");	   
			    
fun print_actFunc tag index stm k =
    if assigns [] stm
    then let fun hasGuard s = case s of StGuard _ => true | StBlock (_,s) => hasGuard s | _ => false
	     fun print_act s dl =
		 case s of
		     StBlock(dl',s) => print_act s (dl @ dl')
		   | StSequence(s1,s2) =>
		     if matchStm hasGuard s1
		     then print_act s2 dl
		     else (newline 0; newline 0; print ("bool act_"^ Int.toString(index) ^" () {"); (*//&*)
			   newline 2; print "/* "; print tag; print " */";
			   (* if null k then () else (newline 2; print "/* temporaries */"); *)
			   List.app (fn d => (newline 2; print_VarDecl d)) k;
			   (*newline 2; print ("struct value new = *(struct value*)(*lookup(s);"); //&*)
			   print_Statement 2 s "state."; (*//&*)
			   newline 2; print ("return 1;"); (*//&*)
			   newline 1; print ("error: return 0;");
			   newline 0; print "}")
		   | _ => ()
	 in	    Array.update(!ACT,index,"act_"^Int.toString(index));  
		    print_act stm []
	 end
    else ();

fun print_transFunc a rk =
    (newline 0; print "\n/* -- predicates and actions */";
     newline 0; print "/* by default, pre_t* returns true and act_t* is the identity function */";
     newline 0; newline 0; print "/* translation table for transition names */";
     newline 0; print ("char *transtable["^ Int.toString(List.length a) ^"] = ");

     XTR := Array.array(List.length a, NONE);
     PRE:= Array.array(List.length a, "0");
     ACT:= Array.array(List.length a, "0");
     
     print_list ("{",",","}") (fn (([st],_,_),_,_,_,_) => print ("\""^ hd (String.fields (fn #":" => true | _ => false) st)^"\"")) a; print ";";
     newline 0; print "\nchar **transitions (int *sz) {";
     newline 0; print "*sz ="; print (Int.toString(List.length a)); print ";";
     newline 0; print "return transtable;";
     newline 0; print "}";
     List.app (fn (i,(([tag],lab,_),_,stm,_,_)) => print_preFunc print_pre_hippo (tag^(case lab of "" => "" | _ => (":"^lab))) i stm [] rk)
	      (ListPair.zip(List.tabulate(List.length a,fn y=>y),a));
     List.app (fn (i,(([tag],lab,sec::tags),_,stm,_,_)) =>
		  (if !TAGS andalso not (null tags) then
		       let open TextIO
			   fun p s = output (stdOut,s)
		       in p tag; p " ->"; List.app (fn g => (p " "; p g)) tags; p "\n"
		       end
		   else
		       ();
		   print_actFunc (tag^(case lab of "" => "" | _ => (":"^lab))) i stm rk))
	      (ListPair.zip(List.tabulate(List.length a,fn y=>y),a)));

(* printing code file *)
fun print_Fiacre (Program ((_,dl,d)::_)) =
    let fun print_dec d =
	    case d of
		FunctionDecl _ => (newline 0; print_FunctionDecl d)
	      | ExternDecl _ => (newline 0; print_ExternDecl d)
	      | TaskDecl _ => (newline 0; print_TaskDecl d)
	      | EventDecl _ => (newline 0; print_EventDecl d)
	      | TypeDecl (s,ty) => print_TypeDecl d
	      | _ => ()
    in newline 0; newline 0; print "static int avail;    /* for sprint_* */";
       newline 0; print "static char tmpbuff[16]; /* for sprint_* */";
       List.app print_dec dl; 
       let val ProcessDecl (a,_,_,_,rk,k,f,_,_,_,j) = d
       in newline 0; newline 0; print "/* Process "; print a; print " */";
	  TEMP := map (fn (v,_,_) => v) (rk @ k);
	  if null rk then () else (newline 0; newline 0; print "/* read-only temporary variables */");
	  List.app (fn d => (newline 0; print_VarDecl d)) rk;
	  newline 0;
	  (* creates value type, compare_value, sprint_value: *)
	  newline 0; print "/* state type */";
	  let val valuetype = RecordType (map (fn (y,ty,_) => (y,ty)) f)
	  in print "\n\n"; print_Type valuetype "value"; print ";";
		  newline 0 ; print "struct value state;"; newline 0; (*//&*)
	     print_TypeDecl (TypeDecl ("value",valuetype))  (* functions *)  
	  end;
	  print_initialFunc f rk;
	  print_transFunc j k
       end;
       newline 0
    end;

(* emits .c iff some data *)
fun emitFiacre (hfile,fdef,writeh) (p as Program ((_,dl,d as ProcessDecl (_,_,_,_,rk,_,f,_,_,_,_))::_)) =
      (print ("/* Created " ^ Date.toString (Date.fromTimeLocal(Time.now ())) ^ " by frac version " ^ version ^ " */");
       newline 0; newline 0; print "#include <stdlib.h>";
       newline 0; print "#include <math.h>";
       newline 0; print "#include <string.h>";
       newline 0; print "#include <stdio.h>";
       newline 0; print "#include <stddef.h>";
       newline 0; print "#include <fiacre.h>";
       newline 0; print "\n/* provider */";
       newline 0;
       print "char *source_info[3] = {\"frac\",\"";
       print version; print "\",\"";
       print (Date.toString (Date.fromTimeLocal(Time.now ()))); print "\"};";
       print "\nchar **sourceinfo () {return source_info;}";
       flush ();
       (* following is disabled as there may be non trivial guards even if no data
          e.g. due to pattern matching used in a communication (PAB example).
	  disabling it should not harm, in any case *)
       if false (* List.null f andalso List.null rk *) then
	   (newline 0; print "/* minimal functions */";
	    newline 0; print "typedef int bool;";
	    newline 0; print "typedef long key;";
	    newline 0; print "char buff[3];";
	    newline 0; print "int sprint_state (int sz, char *buff, key s) {strcpy(buff,\"{}\\0\"); return 2;}";
	    newline 0; print "int sprint_json_state (int sz, char *buff, key s) {strcpy(buff,\"{}\\0\"); return 2;}";
	    newline 0; print "void initial() {}";
	    newline 0; print "char **transitions (int *sz) {*sz =0; return 0;}";
	    newline 0; flush ();
	    OS.FileSys.remove fdef handle _ => ())
       else
	   let val p' as Program ((_,dl',_)::_) = Name.nameTypes p	       
	   in (* Construct the list of types (in order of declaration) *)
	      TYPES := List.mapPartial (fn (TypeDecl(a,ty)) => SOME (a,ty) | _ => NONE) dl';
	      writeh print_Headers p';
	      newline 0; print "\n/* fiacre and private headers */";
	      newline 0; print "#include \"hippo_app_functions.h\""; (*print hfile; print "\""; //&*)
	      print_Fiacre p';
	      newline 0; flush ()
	   end)
      
       handle BYE => raise BYE | _ => failwith "emit c";
   
end; (* structure HippoCEmit *)



(* prints the Hippo Net (data.c) by Prn *)
structure HippoNetEmit = struct

open OutPrint;

val EDGES = ref [];
val EDGES_FORMATED = ref[];
val PRESET = ref [];
val POSTSET = ref [];
val PRESET_FORMATED = ref [];
val POSTSET_FORMATED = ref [];
val SETPRIO = ref [];


val N_PLACES = ref 0;
val N_TRANSITIONS = ref 0;
val N_EDGES = ref 0;
val N_PRE = ref 0;
val N_POST = ref 0;
val N_EXTERNS = ref 0;
val N_ALARMS = ref 0;
val N_TASKS = ref 0;
val N_START = ref 0;
val N_END = ref 0;
val N_PRIORITY_RELATIONS = ref 0;
    
val START_TRANS = ref[];
val END_TRANS = ref[];
val EXTSET_START = ref[]
val EXTSET_END = ref[]

    
fun newline n = if n<0 then () else (print "\n"; repeat n print " ");

fun print_Fiacre (Program ((_,_,ProcessDecl d)::_)) =
    ((* just for testing: *)
    (* print "--- externset ---\n";
     Array.appi (fn (i,NONE) => (printi i; print " => -\n")
		  | (i,SOME (name,var)) => (printi i; print " => "; print name; print ", "; print var; print "\n"))
		(!HippoCEmit.XTR);
     print "---\n\n";*)
    print_Process 0 d)

and get_First (a,_) = a          (*  that's  #1  *)
and get_Second (a,b) = b         (*  that's  #2  *)
and get_Second3 (_,b,_) = b      (*  that's  #2  *)

and get_Third (_,_,c) = c        (*  that's  #3  *)
and get_Thirdint (_,_,c) = c     (*  that's  #3  *)

and format_list [] = []
  |format_list (x::xs) =
   let fun fa (s,i,j,[]) = [(s,i,(j-i))]
	 | fa (s,i,j,x::xs) =
	   if s<>x then
	       (s,i,(j-i))::(fa(x,j,(j+1),xs))
	   else
	       fa(s,i,(j+1),xs)
   in
       (fa (x,0,0,x::xs)
       )
   end
   
and reverse s = implode (foldl op:: [] (explode s))             (* ==  implode (List.rev (explode s))  *)

and compare ((a,b,e),(c,d,f)) = if a=c then true else false	(* ==   a=c   *)	

and qsort(func) =
let
    fun
        sort([]) = [] |
        sort(lhd :: ltl) = sort(List.filter(fn (x) => func(x, lhd))(ltl)) @ [lhd] @ sort(List.filter(fn (x) => not(func(x, lhd)))(ltl))
in sort
end

and printPriorityRelations(t) =	
    let 
		val tt = ListPair.zip(t,List.tabulate(length t,fn i=> i))
		val ln : (string*int)list =
			List.map 
				(fn ((([tag],_,_),_,_,_,_), transitionIndex) => (tag, transitionIndex)) 
				tt
		
		fun transitionWithPriorityRelations (transactions : (Transition*int)list) =
			case (transactions) of
				[] => []
			|	h::t => (case (h) of 
							(((_,_,_),_,_,_,setOfLowerPriorityTransitions),_) => 
								if (length setOfLowerPriorityTransitions)<> 0
									then [h]@(transitionWithPriorityRelations (t))
									else transitionWithPriorityRelations (t)
						)

		val ltp : (Transition*int)list = transitionWithPriorityRelations(tt)

		fun findIndexOfTheTransaction (transactionName : string, setOfTransactionNamesWithIndex : (string*int) list) =
			case (Option.valOf
				(List.find 
					(fn (nt, _) => (nt = transactionName))
					setOfTransactionNamesWithIndex
				)
			) of
				(_, num) => num

		fun numberOfPriorityRelations (setOfTransitions : Transition list) : int =
			case
				setOfTransitions
			of
				[] => 0
			|	((_,_,_),_,_,_,setOfLowerPriorityTransitions)::t => 
					(length setOfLowerPriorityTransitions) + numberOfPriorityRelations (t)
			
		 
		fun prioTrans ((([tag],_,_),_,_,_,setOfLowerPriorityTransitions), transitionIndex: int, ltn) =
			List.app
				(fn (tprio) => (
					SETPRIO := tag :: !SETPRIO;
					print "/* Priority relation between "; 
					print tprio; print " and "; print tag; print " */ "; newline 4;		 
					print "{.lower = &transitionset[";
					print (Int.toString(findIndexOfTheTransaction(tprio, ltn)));
					print "], .higher = &transitionset[";
					print (Int.toString(transitionIndex));
					print "]";
					print "}";
					if transitionIndex <>(length setOfLowerPriorityTransitions)-1 then print"," else (); newline 4))
				setOfLowerPriorityTransitions
    in
		if (numberOfPriorityRelations (t) <> 0)
		then(
			print "PriorityRelation setOfPriorityRelations[NUMBER_OF_PRIORITY_RELATIONS]={\n";
			List.app( fn (tr,n) => prioTrans(tr,n,ln)) ltp ;
			print "};\n"
		)
		else ()
    end
	


and print_Process i (a,_,_,b,_,_,_,_,_,StTo e,c) =
    (* (newline i; print "net process_"; print a; 
	newline i; newline i;*)
    let val tt = ListPair.zip(c,List.tabulate(length c,fn i=> i))
	val ttt = ListPair.zip(tt,List.tabulate(length tt, fn i=> (length tt)-1))
    in
	(print "#include \"hippo_app_data.h\"\n";
	 print "\nTransition *listStartTransitions[NUMBER_OF_START_TRANSITIONS];\nTransition *listEndTransitions[NUMBER_OF_END_TRANSITIONS];\nExtern externset[NUMBER_OF_EXTERNS];\nPlace placeset[NUMBER_OF_PLACES];\nTransition transitionset[NUMBER_OF_TRANSITIONS];\n";
     print "Edge edgeset[NUMBER_OF_EDGES];\nfromPlaceToTransition preset[NUMBER_OF_PRE];\nfromTransitionToPlace postset[NUMBER_OF_POST];\n";
	 print "PriorityRelation setOfPriorityRelations[NUMBER_OF_PRIORITY_RELATIONS];\n\n";

	 Array.app(fn (n) => if n<>"0" then (print "extern bool "; print n; print " ();\n") else ()) (!HippoCEmit.PRE);
	 Array.app(fn (n) => if n<>"0" then (print "extern bool "; print n; print " ();\n") else ()) (!HippoCEmit.ACT);

	 print_StartEnd ();
	 
	 print "\n\nExtern externset[NUMBER_OF_EXTERNS] = {";
	 print_ExternSet ();
	 
	 print "\n};\n\nfromPlaceToTransition preset[NUMBER_OF_PRE] = {\n";
	 newline 4; 
	 print_PreSet (b,c);
	 N_PRE := List.length (!PRESET);
	 
	 print "\nfromTransitionToPlace postset[NUMBER_OF_POST] = {\n";
	 newline 4; 
	 print_PostSet (b,c);
	 N_POST := List.length(!POSTSET);
	 
	 print "\nEdge edgeset[NUMBER_OF_EDGES] = {\n";
	 newline 4;
	 print_EdgeSet (b,c);
	 N_EDGES := List.length(!EDGES);
	 
	 EDGES := List.rev(!EDGES);
	 EDGES_FORMATED := format_list(!EDGES);
	 
	 print "\nPlace placeset[NUMBER_OF_PLACES] = {";
	 newline i; newline 4;
	 print_StatesDecl (b,e);
	 N_PLACES := List.length(b);

	 PRESET := List.rev(!PRESET);
	 POSTSET:= List.rev(!POSTSET);

	 PRESET_FORMATED := format_list(!PRESET);
	 POSTSET_FORMATED := format_list(!POSTSET);
	 	  
	 print "\nTransition transitionset[NUMBER_OF_TRANSITIONS] = {\n";
	 app (print_Transition 4) (ttt);
	 N_TRANSITIONS := List.length(c);

	 N_ALARMS := !N_TRANSITIONS;

	 
	 print "\n\n};\n";newline i;

	 printPriorityRelations (c);
	 N_PRIORITY_RELATIONS := List.length(!SETPRIO)

	)
	
	
    
     (*app (print_prior i) c;*)
				   
    end

and print_StartEnd () =

    (Array.modify(fn NONE => NONE
		   |SOME(name,var) => SOME((reverse name),var))(!HippoCEmit.XTR);
     Array.appi (fn (i,NONE) => ()
		 |(i,SOME (name,var))  => (if  var = "ready" then (START_TRANS := (name,"&transitionset["^Int.toString(i)^"]","1") :: !START_TRANS) else (); if var = "start" then (START_TRANS := (name,"&transitionset["^Int.toString(i)^"]","0") :: !START_TRANS) else (); if var = "terminate" then (END_TRANS := (name,"&transitionset["^Int.toString(i)^"]","0") :: !END_TRANS) else ()))(!HippoCEmit.XTR);

     START_TRANS:= List.rev(!START_TRANS);
     END_TRANS:= List.rev(!END_TRANS);
     START_TRANS := qsort compare (!START_TRANS);
     END_TRANS := qsort compare (!END_TRANS);
     EXTSET_START :=format_list(List.map (fn (a,b,c) => a)(!START_TRANS));
     EXTSET_END := format_list(List.map (fn (a,b,c) => a)(!END_TRANS));
     N_START := List.length(!START_TRANS);
     N_END := List.length(!END_TRANS);
     N_EXTERNS := List.length(!EXTSET_START);
    (* print "\n\n----\n"; List.app(fn (name,num,e) => (print name; print " , "; print num; print " , "; print e; print"\n"))(!START_TRANS);
    print "\n\n----\n"; List.app(fn (name,num,e) => (print name; print " , "; print num; print " , "; print e; print"\n"))(!END_TRANS); print "\n------\n";*)
     
     List.app(fn(name,first,_)=>case (get_Third(List.nth(!START_TRANS, first))) of
				   "1"=> (print "\n/* fiacre event e, cname "; print name; print " */\nint "; print name; print "_ready = 0;\n")
				 | "0"=> (print "\n/* fiacre task t, cname "; print name; print " */\nint "; print name; print "_start = 1;\nint ";
					  print name; print "_terminate=0;\n")
				 |_ => emit_error "error") (!EXTSET_START);
     List.app(fn(name,_,_) => (print "\nextern void wrapper_"; print name; print "(void);"))(!EXTSET_START);
	
	 print "\n\nTransition *listStartTransitions[NUMBER_OF_START_TRANSITIONS] = {\n";
     List.app(fn (name,num,e) => (newline 4; print num; print","))(!START_TRANS);
     print "\n};\n\nTransition *listEndTransitions[NUMBER_OF_END_TRANSITIONS] = {\n\n";
     List.app(fn (name,num,e) => (newline 4;print num; print","))(!END_TRANS);
     print "\n};\n")

    
    (*
    print "\n\n----\n"; List.app(fn (name,num,e) => (print name; print " , "; print (Int.toString(num)); print " , "; print (Int.toString(e)); print"\n"))(!EXTSET_START);
    print "\n\n----\n"; List.app(fn (name,num,e) => (print name; print " , "; print (Int.toString(num)); print " , "; print (Int.toString(e)); print"\n"))(!EXTSET_END); print "\n------\n")*)


and print_ExternSet () =
    let val start_formated = ListPair.zip(!EXTSET_START, List.tabulate(length (!EXTSET_START),fn i=>i))
	val end_formated =  ListPair.zip(!EXTSET_END, List.tabulate(length (!EXTSET_END),fn i=>i))
    in
	List.app(fn ((name,first,nb),i) => ( newline 4; print "{ .idx = "; print (Int.toString(i)); print ","; newline 4;
					     print ".name = \""; print name; print"\","; newline 4;
					     print ".isEvent = "; print(get_Third(List.nth(!START_TRANS, first))); print ","; newline 4;
					     print ".startTransition = &listStartTransitions["; print (Int.toString(first)); print "],"; newline 4;
					     print ".nbStartTransition = "; print (Int.toString(nb)); print ","; newline 4;
					     print ".endTransition = "; if get_Third(List.nth(!START_TRANS, first))= "0" then (print "&listEndTransitions["; print (Int.toString(get_Second3(Option.valOf(List.find(fn (name2,first2,nb2) => name=name2)(!EXTSET_END))))); print "],") else (print "NULL,"); newline 4;
					     print ".nbEndTransition = "; if get_Third(List.nth(!START_TRANS, first))= "0" then ( print (Int.toString(get_Thirdint(Option.valOf(List.find(fn (name2,first2,nb2) => name=name2)(!EXTSET_END))))); print ",") else (print "0,"); newline 4;
					     print ".ready = &"; print name; print "_";
					     if get_Third(List.nth(!START_TRANS, first))= "0" then (print"start,") else (print"ready,"); newline 4;
					     print ".terminate = ";
					     if get_Third(List.nth(!START_TRANS, first))= "0" then (print "&"; print name; print "_terminate,") else (print "NULL,"); newline 4;
					     print ".function = wrapper_"; print name; print "},"
		)) start_formated
    end
					     
     



     
    

and print_EdgeSet(p,t) =
    let val pp = ListPair.zip(p,List.tabulate(length p,fn i=> i))
	val tt = ListPair.zip(t,List.tabulate(length t,fn i=> i))
	fun remove_readmark (src)=
		(if (String.sub(src,String.size src - 1) = #"?") then String.substring(src, 0, (String.size src - 1)) else src)
	fun src_test (src,(n,pl,tag,cpt)) =
		(*if src=pl then*)
		if 
			(remove_readmark(src)=pl)
		then
		(print "/* Edge from "; print src; print " to "; print tag; print " */"; newline 4;
		 print "{.fromPlace = &placeset["; print (Int.toString((get_Second(Option.valOf(List.find (fn (x,_) => x=remove_readmark(src)) pp)))));
		 print "], .toTransition = &transitionset["; print (Int.toString(n));
		 print "], .posEnableState = "; print (Int.toString(cpt)); print " },";  newline 4; newline 4;
		 EDGES:= remove_readmark(src) :: !EDGES
		)
	    (*TODO: trouver un moyen pour ne pas mettre la "," � la derniere ligne du edgeset*)

	    else ()
	fun trans_test (((([tag],_,_),src,_,_,_),n),pl) =
	    List.app (fn (plsrc) => src_test(plsrc))  (ListPair.zip(src,List.tabulate(length src, fn i => (n,pl,tag,i))));
	    
	fun pl_test (pl) =
	   List.app (fn (tr,place) => trans_test(tr,place) ) (ListPair.zip(tt,List.tabulate(length tt, fn i => pl)))
    in
	List.app (fn (pl) => pl_test (pl)) p;
	print "};\n"
    end
    
and print_PreSet(p,t) =
    let val pp = ListPair.zip(p,List.tabulate(length p,fn i=> i))
	val tt = ListPair.zip(t,List.tabulate(length t,fn i=> i))
	fun preset_Trans ((([tag],_,_),src,_,_,_),n) =
	    List.app(fn (pl) => (print "/* Arc from "; 
				 if (String.sub(pl,String.size pl - 1) = #"?") then print (String.substring(pl, 0, (String.size pl - 1))) else print pl;
				 print " to "; print tag; print " */ "; newline 4;				 
				 print "{.fromPlace = &placeset["; print (Int.toString((get_Second(Option.valOf(List.find (fn (x,_) => x=(if (String.sub(pl,String.size pl - 1) = #"?") then (String.substring(pl, 0, (String.size pl - 1))) else pl)) pp)))));
				 print "], .toTransition = &transitionset[";
				 print (Int.toString(n)); print "], "; 
				 print " .read="; 
				 if (String.sub(pl,String.size pl - 1) = #"?") then print "1" else print "0";
				 print "}";
				 if n<>(length tt)-1 then print"," else (); newline 4; PRESET := tag :: !PRESET)) src
    in
	List.app( fn ((tr,n)) => (preset_Trans(tr,n); newline 4)) tt;
	print "};\n"
    end

and get_target st =
    case st of
	StSequence(stm1,stm2) => get_target stm1 @ get_target stm2
      | StTo stl => stl
      | _ => []
    
and print_PostSet(p,t) =
    let val pp = ListPair.zip(p,List.tabulate(length p,fn i=> i))
	val tt = ListPair.zip(t,List.tabulate(length t,fn i=> i))
	fun postset_Trans ((([tag],_,_),_,c,_,_),n) =
	    List.app(fn (pl) => (print "/* Arc from "; print tag; print " to "; print pl; print " */ "; newline 4;
				 print "{.fromTransition = &transitionset[";  print (Int.toString(n));
				 print "], .toPlace = &placeset[";
				 print (Int.toString((get_Second(Option.valOf(List.find (fn (x,_) => x=pl) pp))))); print "] }";
				 if n<>(length tt)-1 then print"," else (); newline 4; POSTSET := tag :: !POSTSET)) (get_target c)
    in
	List.app( fn ((tr,n)) => (postset_Trans(tr,n); newline 4)) tt;
	print "};\n"

    end

    
and index (x, []) = raise Subscript
  | index (x, y::ys) =
    if x = y then 0 else 1 + index (x, ys)


    

    

(* each state in initial state is a marked place *)
(* ##### also prints other states in case no transition refers to it. wise ? ##### *)
and print_StatesDecl (a,b) =
    let val aa = ListPair.zip(a,List.tabulate(length a,fn i=> i))
    in
	List.app (fn (s,n) => (print "{ .idx ="; print (Int.toString (n)) ;
			       print " , .name = \""; print s; print "\", .mark = "; if mem s b then print "1" else print "0";
			       print ", .edge = &edgeset[";
			       print (case (List.find (fn (pl,_,_) => s=pl) (!EDGES_FORMATED)) of
					   NONE => "-1"																		       | SOME (pl,n,nb) => Int.toString(n)
				     );
			       print "], .nbEdges = ";
			       print (case (List.find (fn (pl,_,_) => s=pl) (!EDGES_FORMATED)) of
					  NONE => "0"																		       | SOME (pl,n,nb) => Int.toString(nb)
				     );
			       print " }"; if n <> (length a) -1 then print"," else (); newline 4)) aa;
	print "\n};\n"
    end

and print_Interval (left,right) =
    let fun printb b =
	    print (Int.toString (Real.floor (b * (pow (10.0,!DIGITS)))))
	    handle Overflow => emit_error "interval bound too large"
	val Z = ref false
    in case left of
	   Open b => (print ".minTime = "; printb b)
	 | Closed b => (print ".minTime = "; printb b)
	 | Infinite => print ".minTime = 0";
       print ", ";
       case right of
	   Open b => (print ".maxTime = "; printb b)
	 | Closed b => (print ".maxTime = "; printb b)
	 | Infinite => print ".maxTime = 0"
    end

and print_Interval2 (left,right) =
    let fun printb b =
	    print (Real.toString b)
	    handle Overflow => emit_error "interval bound too large"
	 val Z = ref false
     in case right of
	    Open b => (print ".minTime = "; printb b; print ", "; print ".maxTime = "; printb b)
	  | Closed b => (print ".minTime = "; printb b; print ", "; print ".maxTime = "; printb b)
	  | Infinite => (print ".minTime = 0"; print ", "; print ".maxTime = 0")
     end

and print_Transition i (((([tag],lab,_),src,c,interv,_),n),max) =
    (* tag = nom de la transition,
       lab = etiquette, 
       src = etats source et nature, 
       interv = intervalle temporel, 
       c = "code" de la transition *)
    let (* gets target state *)
	(*fun get_target st =
	    case st of
		StSequence(stm1,stm2) => get_target stm1 @ get_target stm2
	      | StTo stl => stl
	      | _ => []*)
	val tgt = get_target c  (* etats cibles *)
	val (read,inhib,from,tg) =  (* inhib for probes, in case needed *)
	    let val (rd,src') = List.partition (fn p => String.sub(p,String.size p - 1) = #"?") src
		val (inh,fr) = List.partition (fn p => String.sub(p,String.size p - 2) = #"?") src'  (* ends with "?-" *)
	    in (rd,
		inh,
		fr,
		tgt)
	    end
    in newline i; print "{ .idx = "; print (Int.toString (n)); 
       print ", .name = \""; print tag; print "\", ";(* case lab of "" => () | _ => (print ": "; print lab);*)
       print_Interval2 interv;
       print ", .preset = &preset[";
       print (case (List.find (fn (tr,_,_) => tag=tr) (!PRESET_FORMATED)) of
		  NONE => "-1"																		       | SOME (tr,n,nb) => Int.toString(n)
	     );
       print "], .sizePreset = ";
       print (case (List.find (fn (tr,_,_) => tag=tr) (!PRESET_FORMATED)) of
		  NONE => "0"																		       | SOME (tr,n,nb) => Int.toString(nb)
	     );
       print ", .postset = &postset[";
              print (case (List.find (fn (tr,_,_) => tag=tr) (!POSTSET_FORMATED)) of
		  NONE => "-1"																		       | SOME (tr,n,nb) => Int.toString(n)
	     );
       print "], .sizePostset = ";
       print (case (List.find (fn (tr,_,_) => tag=tr) (!POSTSET_FORMATED)) of
		  NONE => "0"																		       | SOME (tr,n,nb) => Int.toString(nb)
	     );
       print ", .pre = "; print (Array.sub(!(HippoCEmit.PRE), n));
       print" , .act = "; print (Array.sub(!(HippoCEmit.ACT), n));
       print" , .call = ";
       case (Array.sub((!HippoCEmit.XTR),n)) of
	   NONE => (print "0")
	 | SOME(name,"start") => (print "&externset[";print (Int.toString(get_First(Option.valOf(List.find(fn (i,(nom,num,nb)) => nom=name)(ListPair.zip(List.tabulate(length (!EXTSET_START),fn i=> i),(!EXTSET_START))))))); print "]")
	 | _ => (print "0");
	   
       print", .alarm = 0"; (*print (Int.toString(n));*)
       print ", .task = 0, .enabled_state = 1, .priority=0 ";
       print "}";  if n <> max then print"," else ()
      (* print_list (" "," ","") (fn p => print (p ^ "1")) read;
       print_list (" "," ","") (fn p => print (p ^ "1")) inhib;
       print_list (" "," ","") print from;
       print " -> ";prin
	print_list (""," ","") print tg*)
       

       
    end

and print_prior i (_,_,_,_,[]) = ()
  | print_prior i (([tg],_,_),_,_,_,prl) = (newline i; print "pr "; print tg; print " > "; print_list (""," ","") print prl)

fun emitFiacre s0 = (print_Fiacre s0; newline 0; flush ()) handle BYE => raise BYE | _ => failwith "emit net";

end; (* structure HippoNetEmit *)



	  
structure HippoDataHEmit = struct

open OutPrint;

fun newline n = if n<0 then () else (print "\n"; repeat n print " ");

	  
fun emitFiacre t =
    (print "#ifndef _APP_DATA_\n#define _APP_DATA_\n\n#include \"data.h\"\n\n/* Static allocation */\n#define NUMBER_OF_PLACES ";
     print (Int.toString(!HippoNetEmit.N_PLACES));
     print "\n#define NUMBER_OF_TRANSITIONS "; print (Int.toString(!HippoNetEmit.N_TRANSITIONS));
     print "\n#define NUMBER_OF_EDGES "; print (Int.toString(!HippoNetEmit.N_EDGES));
     print "\n#define NUMBER_OF_PRE "; print (Int.toString(!HippoNetEmit.N_PRE));
     print "\n#define NUMBER_OF_POST "; print (Int.toString(!HippoNetEmit.N_POST));
     print "\n#define NUMBER_OF_TASKS "; print (Int.toString(!HippoNetEmit.N_EXTERNS));
     print "\n#define NUMBER_OF_ALARMS "; print (Int.toString(!HippoNetEmit.N_ALARMS));
     print "\n#define NUMBER_OF_START_TRANSITIONS "; print (Int.toString(!HippoNetEmit.N_START));
     print "\n#define NUMBER_OF_END_TRANSITIONS "; print (Int.toString(!HippoNetEmit.N_END));
     print "\n#define NUMBER_OF_EXTERNS "; print (Int.toString(!HippoNetEmit.N_EXTERNS));
     print "\n#define NUMBER_OF_PRIORITY_RELATIONS "; print (Int.toString(!HippoNetEmit.N_PRIORITY_RELATIONS));
     (*print "\n\n\nTransition *listStartTransitions[NUMBER_OF_START_TRANSITIONS];\nTransition *listEndTransitions[NUMBER_OF_END_TRANSITIONS];\nExtern externset[NUMBER_OF_EXTERNS];\nPlace placeset[NUMBER_OF_PLACES];\nTransition transitionset[NUMBER_OF_TRANSITIONS];\n";
     print "Edge edgeset[NUMBER_OF_EDGES];\nfromPlaceToTransition preset[NUMBER_OF_PRE];\nfromTransitionToPlace postset[NUMBER_OF_POST];\n";
	 print "PriorityRelation setOfPriorityRelations[NUMBER_OF_PRIORITY_RELATIONS];";*)
	 print"\n\n#endif";

     newline 0; flush()
    )
    handle BYE => raise BYE | _ => failwith "emit c";


end;



structure EmitHippo : EMIT = struct

fun readOnly f = ();
    (*
     let open Posix.FileSys
	 val m = S.flags [S.irusr, S.irgrp, S.iroth]
     in chmod (f, m)
     end;
     *)

(* output is implicitly hippo *)
fun emit t =
    let val dir = !OutPrint.name
	val base = #base(OS.Path.splitBaseExt (#file (OS.Path.splitDirFile dir)))
	(* nom du fichier .net *)
	val fnet = OS.Path.joinDirFile {dir=dir,file=OS.Path.joinBaseExt {base="hippo_app_data",ext=SOME "c"}}
	(* stream fu chichier .net *)
	val snet = TextIO.openOut fnet handle x => emit_error ("cannot open "^fnet)
	val fdat = OS.Path.joinDirFile {dir=dir,file=OS.Path.joinBaseExt {base="hippo_app_functions",ext=SOME "c"}}
	val sdat = TextIO.openOut fdat handle x => emit_error ("cannot open "^fdat)
	val fdef = OS.Path.joinDirFile {dir=dir,file=OS.Path.joinBaseExt {base="hippo_app_functions",ext=SOME "h"}}
	val sdef = TextIO.openOut fdef handle x => emit_error ("cannot open "^fdef)

	val fh =  OS.Path.joinDirFile {dir=dir,file=OS.Path.joinBaseExt {base="hippo_app_data",ext=SOME "h"}}
	val sh =  TextIO.openOut fh handle x => emit_error ("cannot open "^fh)
						
    in printVerbose "::Generating Hippo\n";

       let (* ugly but will improve later *)
	   fun writeh f x =
	       (OutPrint.flush (); OutPrint.name := fdef; OutPrint.FILE := sdef;
		f x;
		OutPrint.flush (); OutPrint.name := fdat; OutPrint.FILE := sdat)
       in OutPrint.name := fdat;
	  OutPrint.FILE := sdat;
	  readOnly fdat;
	  printVerbose "::Writing functions(.c/.h) Files\n";		 
	  HippoCEmit.emitFiacre (base^".h",fdef,writeh) t;
	  readOnly fdef;
	  OutPrint.name := fnet;
	  OutPrint.FILE := snet;
	  printVerbose "::Writing data(.c/.h) File\n";
	  HippoNetEmit.emitFiacre t;
	  readOnly fnet;

	  (**)
	  OutPrint.name:=fh;
	  OutPrint.FILE:=sh;
	  HippoDataHEmit.emitFiacre t;
	  readOnly fh;
	  
	  printVerbose "::Done\n";
	  printVerbose "::Done\n"
       end
    end;

end;


