
(* shared by TtsCEmit and HippoCEmit *)

structure SharedCEmit = struct

open OutPrint;
     
fun newline n = if n<0 then () else (print "\n"; repeat n print " ");

(* hippo only, here because treatment of CVarExp in shared print_Exp requires it *)
val xtr = ref NONE;  (* temporary for recording extern *)
    
(* type environment *)
val TYPES = ref ([] : (string * Type) list);
fun getBoundType s = assoc s (!TYPES);

(* holds temporary variables *)
(* #### could be avoided if temp variables had a specific constructor introduced by simplifier (e.g. TmpVarExp) #### *)
val TEMP = ref ([] : string list);

(* for CALLFUN optimization *)
val IMPURE = ref ([] : string list);
val PURE = ref true;

(* for tmp variable generation *)
(* q, e, r, a, c used in postcompose, with their own numbering; avoid them here *)
(* j, z, used here; i avoided because i386 is typically used in library headers ... *)
val new = let val UNIQUE = ref 0
	  in fn s => let val u = !UNIQUE
		     in UNIQUE := 1 + !UNIQUE;
			let val v = s ^ Int.toString u
			in TEMP := v :: !TEMP;
			   v
			end
		     end
	  end;

fun print_Type ty a =
    let fun printn "" = () | printn a = (print " "; print a)
    in case ty of
	   BoolType => (print "bool"; printn a)
	 | NatType  => (print "unsigned int"; printn a)
	 | CharType => (print "char"; printn a)
	 | UnsignedCharType => (print "unsigned char"; printn a)
	 | ShortType => (print "short"; printn a)
	 | UnsignedShortType => (print "unsigned short"; printn a)
	 | IntType  => (print "int"; printn a)
	 | NamedType nt =>	(case getBoundType nt of
				     EnumType _ => (print "enum "; print nt; printn a)
				   | UnionType _ => (print "struct "; print nt;  printn a) (* unions are records *)
				   | RecordType _ => (print "struct "; print nt;  printn a)
				   | QueueType _ => (print "struct "; print nt;  printn a) (* queues are records *)
				   | ArrayType _ => (print "struct "; print nt;  printn a)
				   | _ => failwith "print_Named_Type")
	 | EnumType cl => (print "enum "; print a; print_list (" {",",","}") print cl)
	 | UnionType ctyl => let val l = List.mapPartial (fn (c,SOME ty) => SOME (c,ty) | _ => NONE) ctyl
			     in print "struct "; print a; print " {";
				newline 2; print "enum "; print_list (" {",",","}") (fn (c,_) => print c) ctyl; print " con;";
				newline 2; print "union {";
				List.app (fn (c,ty) => (newline 4; print_Type ty c; print ";")) l;
				newline 4; print "int none;"; (* now useless ? *)
				newline 2; print "} arg;";
				newline 0; print "}"
			     end
	 | RecordType rl => (print "struct "; print a;
			     print_list (" {","","") (fn (x,y) => (newline 2; print_Type y x; print ";")) rl;
			     newline 0; print "}")
	 | ArrayType (s,t) => (print "struct "; print a; print " {";
			       newline 2; print_Type t "at"; print "["; print (Int.toString s); print "];";
			       newline 0; print "}")			  
	 | QueueType(s,b) => (print "struct "; print a; print " {";
			      newline 2; print "unsigned int len;";
			      newline 2; print_Type b "at"; print "["; print (Int.toString s); print "];";
			      newline 0; print "}")
	 | _ => failwith "print_Type"
    end;

fun print_ExternDecl (ExternDecl (f,argtypes,ty)) =
    (newline 0;
     print "/* C function "; print f; print " */\n";
     print "extern "; print_Type ty f; print_list (" (",",",");") (fn ty => print_Type ty "") argtypes);

(* prints compare_a, sprint_a functions for type a *)
fun print_compareFunc (a,ty) =
    let fun compare i (NamedType n) e1 e2 = (newline i; print ("{int r = compare_"^n^"("^e1^","^e2^"); if (r) return r;}"))
	  | compare i BoolType e1 e2 = (newline i; print ("if (!("^e1^") && ("^e2^")) return -1;");
					newline i; print ("if (("^e1^") && !("^e2^")) return 1;"))
	  | compare i _ e1 e2 = (newline i; print ("if (("^e1^") < ("^e2^")) return -1;");
				 newline i; print ("if (("^e1^") > ("^e2^")) return 1;"))
    in case String.sub(a,0) of
	   #"e" => (* enum *)
	   (newline 0; newline 0; print "static int compare_"; print a; print " (enum "; print a; print " e1, enum "; print a; print " e2) {";
	    newline 2; print "if (e1 < e2) return -1;";
	    newline 2; print "if (e1 > e2) return 1;";
	    newline 2; print "return 0;";
	    newline 0; print "}")
	 | #"u" => (* union *)
	   let val UnionType ctyl = getBoundType a
	       val zary = List.mapPartial (fn (c,NONE) => SOME c | _ => NONE) ctyl
	       val oary = List.mapPartial (fn (c,SOME ty) => SOME (c,ty) | _ => NONE) ctyl
	   in newline 0; print "\nstatic int compare_"; print a; print " (struct "; print a; print " e1, struct "; print a; print " e2) {";
	      newline 2; print ("if (e1.con < e2.con) return -1;");
	      newline 2; print ("if (e1.con > e2.con) return 1;");
	      newline 2; print ("switch (e1.con) {");
	      List.app (fn c => (newline 4; print "case "; print c; print " : {break;}")) zary;
	      List.app (fn (c,ty) => (newline 4; print "case "; print c; print " : {";
				      compare 6 ty ("e1.arg."^c) ("e2.arg."^c);
				      print " break;}")) oary;
	      newline 2; print "}";
	      newline 2; print "return 0;";
	      newline 0; print "}"
	   end
	 | #"q" => (* queue, only present items are compared *)
	   let val QueueType (sz,ety) = getBoundType a
	   in newline 0; print "\nstatic int compare_"; print a; print " (struct "; print a; print " e1, struct "; print a; print " e2) {";
	      compare 2 IntType "e1.len" "e2.len";
	      newline 2; print "{size_t i = 0;";
	      newline 4; print ("for (i=0; i< e1.len; i++) {");
	      compare 6 ety ("e1.at[i]") ("e2.at[i]");
	      newline 4; print "}";
	      newline 2; print "}";
	      newline 2; print "return 0;";
	      newline 0; print "}"
	   end
	 | #"a" => (* array *)
	   let val ArrayType (sz,ety) = getBoundType a
	   in newline 0; print "\nstatic int compare_"; print a; print " (struct "; print a; print " e1, struct "; print a; print " e2) {";
	      newline 2; print "size_t i = 0;";
	      newline 2; print ("for (i=0; i<"^Int.toString sz^"; i++) {");
	      compare 4 ety ("e1.at[i]") ("e2.at[i]");
	      newline 2; print "}";
	      newline 2; print "return 0;";
	      newline 0; print "}"
	   end
	 | #"r" => (* record *)
	   let val RecordType ftyl = getBoundType a
	   in newline 0; print "\nstatic int compare_"; print a; print " (struct "; print a; print " e1, struct "; print a; print " e2) {";
	      List.app (fn (f,ty) => compare 2 ty ("e1."^f) ("e2."^f)) ftyl;
	      newline 2; print "return 0;";
	      newline 0; print "}"
	   end
	 | #"v" => (* value type, a record *)
	   let val RecordType ftyl = ty
	   in newline 0; print "\nint compare_"; print a; print " (struct "; print a; print " *e1, struct "; print a; print " *e2) {";
	      List.app (fn (f,ty) => compare 2 ty ("e1->"^f) ("e2->"^f)) ftyl;
	      newline 2; print "return 0;";
	      newline 0; print "}"
	   end
    end;

(* prints sprint_a function for type a *)
(* almost shared by tts and hippo; specialization by do_sprint_value *)
fun print_sprintFunc do_sprint_value (a,ty) =
     let fun sprint i (NamedType n) e = (case String.sub(n,0) of
					     #"e" => (* enum *) print ("buff = sprint_"^n^" (buff,"^e^",pmap);")
					   | _ => print ("buff = sprint_"^n^" (buff,&"^e^",pmap);"))
	   | sprint i (BoolType) e = print ("if ("^e^") {PUTN(4,pmap->n,\"true\");} else {PUTN(5,pmap->n,\"false\");}")
	   | sprint i (CharType) e = print ("PUTV(\"%hhd\","^e^");")
	   | sprint i (UnsignedCharType) e = print ("PUTV(\"%hhu\","^e^");")
	   | sprint i (ShortType) e = print ("PUTV(\"%hd\","^e^");")
	   | sprint i (UnsignedShortType) e = print ("PUTV(\"%hu\","^e^");")
	   | sprint i (NatType) e = print ("PUTV(\"%u\","^e^");")
	   | sprint i _ e = print ("PUTV(\"%d\","^e^");")
     in case String.sub(a,0) of
	    #"e" => (* enum *)
	    let val EnumType cl = ty
	    in newline 0; newline 0; print "static char *sprint_"; print a; print " (char *buff, enum "; print a; print " e, const struct pmap *pmap) {";
	       newline 2; print "switch (e) {";
	       List.app (fn c => (newline 4; print "case "; print c;
				  print " : {PUTN("; printi (String.size c); print ",pmap->n,\""; print c; print "\"); break;}"))
			cl;
	       newline 2; print "}";
	       newline 2; print "return buff;";
	       newline 0; print "}"
	    end
	  | #"u" => (* union *)
	    let val UnionType ctyl = getBoundType a
		val zary = List.mapPartial (fn (c,NONE) => SOME c | _ => NONE) ctyl
		val oary = List.mapPartial (fn (c,SOME ty) => SOME (c,ty) | _ => NONE) ctyl
	    in newline 0; newline 0; print "static char *sprint_"; print a; print " (char *buff, struct "; print a; print " *e, const struct pmap *pmap) {";
	       newline 2; print "switch (e->con) {";
	       List.app (fn (c,SOME ty) => (newline 4; print "case "; print c; print " : {";
					    newline 6; print "PUTS(pmap->c.l);";
					    newline 6; print "PUTN("; printi (String.size c); print ",pmap->n,\""; print c; print "\");";
					    newline 6; print "PUTS(pmap->c.b);";
					    newline 6; sprint 6 ty ("e->arg."^c); print "";
					    newline 6; print "PUTS(pmap->c.r);";
					    newline 6; print "break;}")
			  | (c,_) => (newline 4; print "case "; print c;
				      print " : {PUTN("; printi (String.size c); print ",pmap->n,\""; print c; print "\"); break;} ")
			) ctyl;
	       newline 2; print "}";
	       newline 2; print "return buff;";
	       newline 0; print "}"
	    end
	  | #"q" => (* queue *)
	    let val QueueType(sz,ety) = ty
		val q = new "q"
	    in newline 0; newline 0; print "static char *sprint_"; print a; print " (char *buff, struct "; print a; print " *e, const struct pmap *pmap) {";
	       newline 2; print "int "; print q; print ";";
	       newline 2; print "PUTS(pmap->q.l);";
	       newline 2; print ("for ("^q^"=0; "^q^"<e->len; "^q^"++) {");
	       newline 4; print ("if ("^q^") {PUTS(pmap->q.s);}");
	       newline 4; sprint 8 ety ("e->at["^q^"]");
	       newline 2; print "}";
	       newline 2; print "PUTS(pmap->q.r);";
	       newline 2; print "return buff;";
	       newline 0; print "}"
	     end
	  | #"a" => (* array *)
	    let val ArrayType(sz,ety) = ty
		val j = new "j"
	    in newline 0; newline 0; print "static char *sprint_"; print a; print " (char *buff, struct "; print a; print " *e, const struct pmap *pmap) {";
	       newline 2; print "int "; print j; print ";";
	       newline 2; print "PUTS(pmap->a.l);"; 
	       newline 2; print ("for ("^j^"=0; "^j^"<"^Int.toString sz^"; "^j^"++) {");
	       newline 4; print ("if ("^j^") {PUTS(pmap->a.s);}");
	       newline 4; sprint 4 ety ("e->at["^j^"]");
	       newline 2; print "}";
	       newline 2; print "PUTS(pmap->a.r);"; 
	       newline 2; print "return buff;";
	       newline 0; print "}"
	    end
	  | #"r" => (* record *)
	    let val RecordType ftyl = ty
		fun spr [(f,ty)] = (newline 2; print ("PUTN(" ^ Int.toString (String.size f) ^ ",pmap->n,\"" ^ f ^ "\"); "); print "PUTS(pmap->r.b); ";
				    sprint 2 ty ("e->" ^ f))
		  | spr (h::(t as _::_)) = (spr [h]; newline 2; print "PUTS(pmap->r.s);"; spr t)
	    in newline 0; newline 0; print "static char *sprint_"; print a; print " (char *buff, struct "; print a; print " *e, const struct pmap *pmap) {";
	       newline 2; print "PUTS(pmap->r.l);";
	       spr ftyl;
	       newline 2; print "PUTS(pmap->r.r);";
	       newline 2; print "return buff;";
	       newline 0; print "}"
	     end
	  | #"v" => (* value type, a record; also initialize avail, avoid surrounding quotes *)
	    let val RecordType ftyl = ty
		fun spr [] = ()
		  | spr [(f,ty)] = (newline 2; print ("PUTN(" ^ Int.toString (String.size f) ^ ",pmap->n,\"" ^ f ^ "\"); "); print "PUTS(pmap->r.b); ";
				    sprint 2 ty ("v->" ^ f))
		  | spr (h::(t as _::_)) = (spr [h]; newline 2; print "PUTS(pmap->r.s);"; spr t)
	    in do_sprint_value spr ftyl
	    end
     end;

fun print_VarDecl (st,ty,_) =
    (print_Type ty st; print ";");

(* returns true if statement or expression updates some state variable *)
fun assigns dl st =
    case st of
	StGuard e => assignse dl e
      | StSequence(stm1,stm2) => assigns dl stm1 orelse assigns dl stm2
      | StAssign(p,exp) => assignsp dl p orelse assignse dl exp
      | StWhile (e,s) => assignse dl e orelse assigns dl s
      | StForeach (v,s) => assignse dl v orelse assigns dl s
      | StCond (e,s1,s2) => assignse dl e orelse assigns dl s1 orelse assigns dl s2
      | StBlock (_,s) => assigns dl s
      | _ => false
	     
(* CVarExp only possible for hippo *)
and assignsp dl p =
    case p of
	VarExp (v,_) => (case List.find (fn (v',_) => v=v') dl of SOME _ => false | _ => not (mem v (!TEMP)))
      | CVarExp (v,_) => true (* special case *)
      | ArrayAccessExp(a,i,_) => assignsp dl a orelse assignse dl i
      | RecordAccessExp(r,_,_) => assignsp dl r
				  
and assignse dl e =
    matchExp (fn LetExp (s,e) => assigns dl s | _ => false) e;
  
(* prints statements in a act_* (without StGuard) *)
fun print_Statement i stm stt =
    case stm of
        StSequence(stm1,stm2) => (print_Statement i stm1 stt; print_Statement i stm2 stt)
      | StCond(e,s1,StNull) => (newline i; print "if ("; print_Exp e stt; print ") {";
				print_Statement (i+2) s1 stt;
				newline i; print "}")
      | StCond(e,StNull,s2) => (newline i; print "if (!("; print_Exp e stt; print ")) {";
				print_Statement (i+2) s2 stt;
				newline i; print "}")
      | StCond(e,s1,s2) => (newline i; print "if ("; print_Exp e stt; print ") {";
			    print_Statement (i+2) s1 stt;
			    newline i; print "} else {";
			    print_Statement (i+2) s2 stt;
			    newline i; print "}")
      | StWhile (e,s) => (newline i; print "while ("; print_Exp e stt; print ") {";
			  print_Statement (i+2) s stt;
			  newline i; print "}")
      | StForeach (e as VarExp(v,IntervalType(fr,to)),s) => (newline i; print "for (";
							     print_Exp e stt; print "="; printi fr; print ";";
							     print_Exp e stt; print "<="; printi to; print ";";
							     print_Exp e stt; print "++) {";
							     print_Statement (i+2) s stt;
							     newline i; print "}";
							     (* so that variable has a legal value on loop exit if not temporary : *)
							     if mem v (!TEMP) then () else (newline i; print_Exp e stt; print "--;"))

      | StAssign(acc,exp) =>
	(* expansion must be applied recursively. e.g.  a := b when a : array 4 of array 8 of ... *)
	if acc=exp then () else
	(case exp of
	     LetExp (stm,e) => (print_Statement i stm stt; print_Statement i (StAssign(acc,e)) stt)
	   | RecordExp (fle,RecordType ftyl) =>
	     List.app (fn (f,e) => print_Statement i (StAssign(RecordAccessExp(acc,f,assoc f ftyl),e)) stt) fle
	   | ArrayExp (el,ArrayType (sz,ety)) =>
	     (List.tabulate (sz,fn j => print_Statement i (StAssign(ArrayAccessExp(acc,IntExp(j,IntType),ety),List.nth(el,j))) stt); ())
	   | QueueExp (el,QueueType(sz,ety)) =>
	     let val len = length el
		 val aty = ArrayType(sz,ety)
	     in print_Statement i (StAssign(RecordAccessExp(acc,"len",IntType),IntExp(len,IntType))) stt;
		List.app (fn j => let val s = StAssign(ArrayAccessExp(acc,IntExp(j,IntType),ety),List.nth(el,j))
				  in print_Statement i s stt
				  end)
			 (List.tabulate(len,fn k => k))
	     end
	   | BinopExp(ENQUEUE,QueueType(sz,ety),q,e) =>
	     (PURE := false; newline i; print "ENQUEUE("; print_Exp acc stt; print ","; print_Exp q stt; print ","; print_Exp e stt; print ","; printi sz; print "); ")
	   | BinopExp(APPEND,QueueType(sz,ety),q,e) => (* assign result then modify *)
	     (PURE := false; newline i; print "APPEND("; print_Exp acc stt; print ","; print_Exp q stt; print ","; print_Exp e stt; print ","; printi sz; print "); ")
	   | UnopExp(DEQUEUE,QueueType(sz,ety),q) =>
	     (PURE := false; newline i; print "DEQUEUE("; print_Exp acc stt; print ","; print_Exp q stt; print "); ")
	   | ConExp (c,ty) =>
	     (newline i; print_Exp acc stt; print ".con = "; print c; print ";")
	   | ConstrExp (c,e,UnionType ctyl) =>
	     (newline i; print_Exp acc stt; print ".con = "; print c; print "; ";
	      print_Statement i (StAssign(RecordAccessExp(acc,"arg." ^ c,let val SOME ty = assoc c ctyl in ty end),e)) stt)
 	   | _ => (newline i; print_Exp acc stt; print "="; print_Exp exp stt; print"; "))
      | StGuard exp => ()
      | StBlock (dl,ss) =>
	(newline i; print "{";
	 List.app (fn (s,ty) => (TEMP := s :: !TEMP; print_Type ty s; print "; ")) dl;
	 print_Statement i ss stt;
	 newline i; print "}")
      | StReturn exp => (newline i; print "return "; print_Exp exp ""; print ";")
      | StFailwith (MatchError f) => (PURE := false; newline i; print "MATCHERROR("; printi f; print ");")
      | _ => ()

(* prints guards or rhs of assignements or initializer components *)
(* CVarExp only possible for hippo *)
and print_Exp e stt =
    case e of
	VarExp (a,_)            => (if mem a (!TEMP) then () else print stt; print a)
      | CVarExp (a,_)           => let fun chopList f [] accu = (String.implode accu,"")
					 | chopList f (l as (h::t)) accu =
					   if f h then (String.implode accu,String.implode t) else chopList f t (h::accu)
				       val (tail,cname) = chopList (fn #"_" => true | _ => false) (List.rev (String.explode a)) []
				   in if mem tail ["ready","start","terminate"] then xtr := SOME (cname,tail) else ();
				      print a
				   end
      | RecordAccessExp (a,b,_) => (print_Exp a stt; print "."; print b)
      | ArrayAccessExp (a,b,_)  => (print_Exp a stt; print ".at["; print_Exp b stt; print "]")
      | IntExp (n,_)            => printi n
      | BoolExp b               => if b then print "1" else print "0" 
      | EnumExp (e,_)           => print e
      | DestrExp (c,e)          => (print_Exp e stt; print (".arg." ^ c))
      | MatchesExp (c,e)        => (print "("; print_Exp e stt; print ".con == "; print c; print ")")
      | InfixExp(x,ty,y,z) 	=>
	(case x of
	     ADD => print_arith_infix ty (y,z) "ADD" "+" stt
	   | SUB => print_arith_infix ty (y,z) "SUB" "-" stt
	   | MUL => print_arith_infix ty (y,z) "MUL" "*" stt
	   | DIV => print_arith_infix ty (y,z) "DIV" "/" stt
	   | MOD => print_arith_infix ty (y,z) "MOD" "%" stt
	   | EQ  => (case ty of
			 NamedType n => (print "(! compare_"; print n; print "("; print_Exp y stt; print ","; print_Exp z stt; print "))")
		       | _ => (print "("; print_Exp y stt; print " == "; print_Exp z stt; print ")"))
	   | NE  => (* if matches (y,z) then print "0" else *) (print "("; print "!"; print_Exp (InfixExp (EQ,ty,y,z)) stt; print ")")
	   | _ => let val f = case x of AND => "&&" | OR  => "||" | LE  => "<=" | LT  => "<" | GT  => ">" | GE  => ">="
		  in print "("; print_Exp y stt; print " "; print f; print " "; print_Exp z stt; print ")"
		  end)
      | UnopExp(a,ty,b)         =>
	(case a of
	     MINUS  => print_arith_unop ty b "MINUS" "-" stt
	   | PLUS   => print_Exp b stt
	   | COERCE => print_arith_unop ty b "COERCE" "" stt
	   | NOT    => (case b of
			    UnopExp(NOT,_,b') => print_Exp b' stt
			  | _ => (print "("; print "!"; print_Exp b stt; print ")"))
	   | ASSERT i => (PURE := false; print "(("; print_Exp b stt; print ") ? 1 : FCR_ASSERT("; printi i; print "))")
	   | _ =>
	     let val QueueType(sz,_) = ty
	     in case a of
		    EMPTY  => (print "EMPTY("; print_Exp b stt; print ")")
		  | FULL   => (print "FULL("; print_Exp b stt; print ","; printi sz; print ")")
		  | LENGTH => (print "LENGTH("; print_Exp b stt; print ")")
		  | FIRST  => (PURE := false; print "FIRST("; print_Exp b stt; print ")")
		  | _ => failwith "unexpected queue unopexp" (* dequeue *)
	     end)
      | CondExp(c,e1,e2)        => (print "(("; print_Exp c stt; print ") ? ("; print_Exp e1 stt; print ") : ("; print_Exp e2 stt; print "))")
      | LetExp(s,e)             => (print "({"; print_Statement (~1000) s stt; print_Exp e stt; print ";})")
      | AppExp (f,el,_,ty)       => (case List.find (fn g => f=g) (!IMPURE) of
	  SOME _ => (PURE := false; print "CALLFUN("; print_Type ty ", "; print ("fcr_"^f); print_list ("(",",","))") (fn e => print_Exp e stt) el)
	| _ => (print ("fcr_"^f); print_list ("(",",",")") (fn e => print_Exp e stt) el) )
      | CallExp (f,el,_,ty)      => (* f assumed impure ... *)
	(PURE := false; print "CALLFUN("; print_Type ty ", ";  print f; print_list ("(",",","))") (fn e => print_Exp e stt) el)
      (* the following cases should not appear; normalized in ? *)
      | BinopExp(x,ty,y,z)	=> failwith "unexpected binop expression in print_Exp"
      | RecordExp _             => failwith "unexpected record expression in print_Exp"
      | ArrayExp _              => failwith "unexpected array expression in print_Exp"
      | QueueExp _              => failwith "unexpected queue expression in print_Exp"
      | ConExp (c,_)            => failwith ("unexpected 0-ary construction ("^c^") in print_Exp")
      | ConstrExp (c,_,_)       => failwith ("unexpected 1-ary construction ("^c^" ...) in print_Exp")
      | _                       => failwith "unexpected expression in print_Exp"

and print_arith_infix ty (y,z) f ope stt =
    case ty of
	IntervalType(l,h) => (PURE := false; print ("I" ^ f); print "("; print_Exp y stt; print ","; print_Exp z stt; print ","; printi l; print ","; printi h; print ")")
      | NatType => (PURE := false; print ("N" ^ f); print "("; print_Exp y stt; print ","; print_Exp z stt; print ")")
      | _ => (print "("; print_Exp y stt; print " "; print ope; print " "; print_Exp z stt; print ")")

and print_arith_unop ty x f ope stt =
    case ty of
	IntervalType(l,h) => (PURE := false; print ("I" ^ f); print "("; print_Exp x stt; print ","; printi l; print ","; printi h; print ")")
      | NatType => (PURE := false; print ("N" ^ f); print "("; print_Exp x stt; print ")")
      | _ => (print "("; print ope; print_Exp x stt; print ")");

fun print_FunctionDecl (FunctionDecl (f,args,ty,vars,body)) =
    (print "\n/* Fiacre function "; print f; print " */";
     newline 0; print_Type ty ("fcr_"^f);
     print_list (" (",", ",") {") (fn (x,ty) => print_Type ty x) args;
     if null vars then () else (newline 4; print_list ("","; ","; ") (fn (x,ty,_) => print_Type ty x) vars);
     (* optimization: do not generate error line, nor CALLFUN in calls, if no macro is required in body *)
     PURE := true; print_Statement 4 body "";
     if !PURE then
	 ()
     else
	 (IMPURE := f :: !IMPURE; newline 3;
	  print "{"; print_Type ty "bogus; error: return bogus;}");
     newline 0; print "}");

(* initializers have been statically evaluated, hence can only be "value" expressions, without variables *)
fun print_Initializer st v stt =
    case v of
	BoolExp _        => (newline 2; print stt; print st; print "="; print_Exp v ""; print";")
      | IntExp _         => (newline 2; print stt; print st; print "="; print_Exp v ""; print";")
      | EnumExp (a,_)    => (newline 2; print stt; print st; print "="; print a; print";")
      | RecordExp (fl,_) => List.app (fn (f,l) => (print_Initializer (st^"."^f) l stt)) fl
      | ArrayExp (el,_)  => List.app (fn (x,y) => print_Initializer (st^".at["^ Int.toString x  ^"]") y stt) 
				     (ListPair.zip(List.tabulate(List.length el,fn z => z),el)) 
      | QueueExp (el,QueueType (z,ty)) =>
	let val len = List.length el
	in newline 2; print stt; print st; print ".len="; printi (length el); print";";
	   List.app (fn (x,y) => print_Initializer (st^".at["^ Int.toString x  ^"]") y stt) 
		    (ListPair.zip(List.tabulate(len,fn z => z),el))
	end
      | ConExp (c,ty) =>
	(newline 2; print stt; print st; print ".con="; print c; print";")
      | ConstrExp (c,e,_) =>
	(newline 2; print stt; print st; print ".con="; print c; print";";
	 print_Initializer (st ^ ".arg." ^ c) e stt)
      | _ => (newline 2; print stt; print st; print "="; print_Exp v ""; print";")
      (* formerly:
      | AppExp _ => (newline 0; print stt; print st; print "="; print_Exp v ""; print";")
      | CallExp _ => (newline 0; print stt; print st; print "="; print_Exp v ""; print";")
      | _ => failwith "unexpected initializer"
      but calls under operators such as 100/f(4000) are legal, though not statically evaluated.
      Alternative would be to evaluate function calls statically too, but this
      requires a true fiacre interpreter ...
       *);

(* almost shared by tts and hippo; specialization by do_print_pre *)
fun print_preFunc do_print_pre tag index stm dl k =
    case stm of
	StBlock(dl',s) => print_preFunc do_print_pre tag index s (dl @ dl') k
      | StSequence(s,_) => print_preFunc do_print_pre tag index s dl k
      | StGuard e =>
	let fun print_unroll_Exp e prf = (* user friendly *)
		case e of
		    LetExp (s,e') => (print_Statement 2 s prf; print_unroll_Exp e' prf)
		  | InfixExp (AND,_,e1,e2) => (print_unroll_Exp e1 prf; print_unroll_Exp e2 prf)
		  | UnopExp(NOT,_,UnopExp(NOT,_,e)) => print_unroll_Exp e prf
		  | UnopExp(NOT,_,e) => (newline 2; print "if ("; print_Exp e prf; print ") return 0;")
		  | UnopExp(ASSERT i,_,e) => (newline 2; print "if (!("; print_Exp e prf; print ")) FCR_ASSERT("; printi i; print ");")
		  | _ => (newline 2; print "if (!("; print_Exp e prf; print ")) return 0;")
	    fun print_guard dl e prf =
		if null dl
		then print_unroll_Exp e prf
		else (newline 2;
		      print "{"; List.app (fn (s,ty) => (TEMP := s :: !TEMP; newline 4; print_Type ty s; print "; ")) dl;
		      print_unroll_Exp e prf;
		      newline 2; print "}")
	in do_print_pre print_guard e tag index dl k
	end
      | _ => (* no guard *) ()

end;  (* SharedCEmit *)
	  

